require 'spec_helper'

describe GroupReportController do
  integrate_views

  def fixture(name)
    File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/#{name}.yaml"))
  end

  before :each do
    @foo_report = Report.create_from_yaml(fixture('foo'))
    @bar_report = Report.create_from_yaml(fixture('bar'))
    @baz_report = Report.create_from_yaml(fixture('baz'))

    @group = NodeGroup.generate!
    @group.nodes = [@foo_report.node, @bar_report.node, @baz_report.node]
    @group.save!
  end

  describe 'without baseline data' do
    describe '#individual' do
      it "should not make any summary data" do
        get :individual, :id => @group.id, :date => @foo_report.time.to_date.to_s(:db)

        response.should be_success

        assigns[:date].should == @foo_report.time.to_date
        assigns[:summary_data].should == {:unreviewed_nodes => 'N/A', :unreviewed => 'N/A', :accepted => 'N/A', :rejected => 'N/A'}
        assigns[:unreviewed_changes_on_other_days].should == 0
        assigns[:other_days].should == []
      end

      it "should default to today if no date is specified" do
        get :individual, :id => @group.id

        response.should be_success

        assigns[:date].should == Time.zone.now.to_date
      end
    end

    describe '#common'do
      it "should redirect to the #individual action" do
        get :common, :id => @group.id, :date => @foo_report.time.to_date.to_s(:db)

        response.should redirect_to :action => :individual, :id => @group.id, :date => @foo_report.time.to_date.to_s(:db)
      end
    end

    describe '#failures'do
      it "should redirect to the #common action" do
        get :failures, :id => @group.id, :date => @foo_report.time.to_date.to_s(:db)

        response.should redirect_to :action => :common, :id => @group.id, :date => @foo_report.time.to_date.to_s(:db)
      end
    end

    describe '#export' do
      it "should not die when rendering a CSV" do
        get :export, :id => @group.id, :date => @foo_report.time.to_date.to_s(:db), :format => 'csv'
        response.should be_success
      end
    end
  end

  describe 'with baseline data' do
    before :each do
      Summarizer.stubs(:delay).returns(Summarizer)

      @foo_baseline_report = Factory(:inspect_report, :host => 'foo')
      @bar_baseline_report = Factory(:inspect_report, :host => 'bar')
      @baz_baseline_report = Factory(:inspect_report, :host => 'baz')

      @foo_baseline = NodeBaseline.create_from_report!(@foo_baseline_report, @foo_report.time)
      @bar_baseline = NodeBaseline.create_from_report!(@bar_baseline_report, @bar_report.time)
      @baz_baseline = NodeBaseline.create_from_report!(@baz_baseline_report, @baz_report.time)

      Summarizer.queue_diff_summarization
    end

    describe "#common" do
      it "should populate the summary data and common changes" do
        get :common, :id => @group.id, :date => @foo_report.time.to_date.to_s(:db)

        response.should be_success

        assigns[:date].should == @foo_report.time.to_date
        assigns[:summary_data].should be_present
        assigns[:unreviewed_changes_on_other_days].should == 0
        assigns[:other_days].should == []
      end

      it "should default to today if no date is specified" do
        get :common, :id => @group.id

        assigns[:date].should == Time.zone.now.to_date
        assigns[:summary_data].should == {:unreviewed_nodes => 'N/A', :unreviewed => 'N/A', :accepted => 'N/A', :rejected => 'N/A'}
        assigns[:unreviewed_changes_on_other_days].should == BaselineDifference.count
        assigns[:other_days].should be_present
      end

      it "should redirect to the individual differences page if there are no common changes" do
        get :common, :id => @group.id, :date => (@foo_report.time.to_date - 1.day).to_s(:db)

        response.should redirect_to :action => :individual, :id => @group.id, :date => (@foo_report.time.to_date - 1.day).to_s(:db)
      end
    end

    describe "#failures" do
      before do
        @failed_resource = @foo_report.resource_statuses.first
        @failed_resource.update_attributes!(:failed => true)
      end

      it "should populate the failures array" do
        get :failures, :id => @group.id, :date => @foo_report.time.to_date.to_s(:db)

        response.should be_success

        assigns[:date].should == @foo_report.time.to_date
        assigns[:failures].map(&:attributes).should =~ [
          {
            "name" => "foo",
            "failures" => "1",
            "node_id" => @foo_report.node_id
          }
        ]
      end

      it "should default to today if no date is specified" do
        get :failures, :id => @group.id

        assigns[:date].should == Time.zone.now.to_date
        assigns[:failures].should_not be_present
      end

      it "should redirect to the common differences page if there are no failures" do
        get :failures, :id => @group.id, :date => (@foo_report.time.to_date - 1.day).to_s(:db)

        response.should redirect_to :action => :common, :id => @group.id, :date => (@foo_report.time.to_date - 1.day).to_s(:db)
      end
    end

    describe '#export' do
      it "should not die when rendering a CSV" do
        get :export, :id => @group.id, :date => @foo_report.time.to_date.to_s, :format => 'csv'
        response.should be_success
      end
    end

    describe "#mark_as_reviewed" do
      it "should mark the corresponding set of differences as 'accepted'" do
        foo_diff = @foo_baseline.differences.find_by_resource_name('File[/etc/sudoers]')
        bar_diff = @bar_baseline.differences.find_by_resource_name('File[/etc/sudoers]')

        get :mark_as_reviewed, :id => @group.id, :date => @foo_report.time.strftime('%F'), :resource => 'File[/etc/sudoers]', :digest => foo_diff.resource_status_digest, :state => 'accepted'

        response.should be_success
        BaselineDifference.find_all_by_accepted(true).should =~ [foo_diff, bar_diff]
        BaselineDifference.find_all_by_accepted(false).should_not be_present
      end

      it "should mark the corresponding set of differences as 'rejected'" do
        foo_diff = @foo_baseline.differences.find_by_resource_name('File[/etc/sudoers]')
        bar_diff = @bar_baseline.differences.find_by_resource_name('File[/etc/sudoers]')

        get :mark_as_reviewed, :id => @group.id, :date => @foo_report.time.strftime('%F'), :resource => 'File[/etc/sudoers]', :digest => foo_diff.resource_status_digest, :state => 'rejected'

        response.should be_success
        BaselineDifference.find_all_by_accepted(true).should_not be_present
        BaselineDifference.find_all_by_accepted(false).should =~ [foo_diff, bar_diff]
      end

      it "should fail if there is no resource specified" do
        foo_diff = @foo_baseline.differences.find_by_resource_name('File[/etc/sudoers]')
        expect {
          get :mark_as_reviewed, :id => @group.id, :date => @foo_report.time.strftime('%F'), :digest => foo_diff.resource_status_digest, :state => 'rejected'
        }.to raise_error(ActiveRecord::RecordNotFound, /Could not find any BaselineDifference/)
      end

      it "should fail if there are no specified differences" do
        expect {
          get :mark_as_reviewed, :id => @group.id, :date => @foo_report.time.strftime('%F'), :resource => 'File[/etc/sudoers]', :digest => 'not a real digest', :state => 'rejected'
        }.to raise_error(ActiveRecord::RecordNotFound, /Could not find any BaselineDifference/)

        expect {
          get :mark_as_reviewed, :id => @group.id, :date => @foo_report.time.strftime('%F'), :resource => 'File[/etc/sudoers]', :digest => 'also not a real digest', :state => 'accepted'
        }.to raise_error(ActiveRecord::RecordNotFound, /Could not find any BaselineDifference/)

        BaselineDifference.find_all_by_accepted(true).should_not be_present
        BaselineDifference.find_all_by_accepted(false).should_not be_present
      end

      it "should fail if setting state to anything but 'accepted' or 'rejected'" do
        get :mark_as_reviewed, :id => @group.id, :date => @foo_report.time.strftime('%F'), :resource => 'File[/etc/sudoers]', :digest => 'ALSO not a real digest', :state => 'awesome'

        response.should_not be_success
        response.code.should == '400'
        response.body.should =~ /'awesome' is not an acceptable state for a BaselineDifference/

        BaselineDifference.find_all_by_accepted(true).should_not be_present
        BaselineDifference.find_all_by_accepted(false).should_not be_present
      end

      describe 'all changes in a contiguous block of differences' do
        before do
          @foo_baselines = (1..5).map do |i|
            NodeBaseline.create_from_report!(@foo_baseline_report, @foo_report.time - i.days).tap do |b|
              b.differences = @foo_baseline.differences.map(&:clone)
              b.save!
            end
          end.unshift(@foo_baseline)

          @bar_baselines = (1..5).map do |i|
            NodeBaseline.create_from_report!(@bar_baseline_report, @bar_report.time - i.days).tap do |b|
              b.differences = @bar_baseline.differences.map(&:clone)
              b.save!
            end
          end.unshift(@bar_baseline)

          @expected = (@foo_baselines + @bar_baselines).map do |bl|
            bl.differences.find_by_resource_name('File[/etc/sudoers]')
          end
        end

        it "should be marked as 'accepted' when state is 'accepted'" do
          foo_diff = @foo_baseline.differences.find_by_resource_name('File[/etc/sudoers]')

          get :mark_as_reviewed, :id => @group.id, :date => @foo_report.time.strftime('%F'), :resource => 'File[/etc/sudoers]', :digest => foo_diff.resource_status_digest, :state => 'accepted'

          response.should be_success
          BaselineDifference.find_all_by_accepted(true).should =~ @expected
          BaselineDifference.find_all_by_accepted(false).should_not be_present
        end

        it "should be marked as 'rejected' when state is 'rejected" do
          foo_diff = @foo_baseline.differences.find_by_resource_name('File[/etc/sudoers]')

          get :mark_as_reviewed, :id => @group.id, :date => @foo_report.time.strftime('%F'), :resource => 'File[/etc/sudoers]', :digest => foo_diff.resource_status_digest, :state => 'rejected'

          response.should be_success
          BaselineDifference.find_all_by_accepted(true).should_not be_present
          BaselineDifference.find_all_by_accepted(false).should =~ @expected
        end

        describe 'where the discontinuity is not the beginning of time' do
          before do
            @foo_baselines[-2].differences.find_by_resource_name('File[/etc/sudoers]').update_attributes! :resource_status_digest => ''
            @bar_baselines[-3].differences.find_by_resource_name('File[/etc/sudoers]').update_attributes! :resource_status_digest => ''

            @expected = (@foo_baselines[0...-2] + @bar_baselines[0...-3]).map do |bl|
              bl.differences.find_by_resource_name('File[/etc/sudoers]')
            end
          end

          it "should mark only the changes in a contiguous block of differences as 'accepted'" do
              foo_diff = @foo_baseline.differences.find_by_resource_name('File[/etc/sudoers]')

              get :mark_as_reviewed, :id => @group.id, :date => @foo_report.time.strftime('%F'), :resource => 'File[/etc/sudoers]', :digest => foo_diff.resource_status_digest, :state => 'accepted'

              response.should be_success
              BaselineDifference.find_all_by_accepted(true).should =~ @expected
              BaselineDifference.find_all_by_accepted(false).should_not be_present
          end

          it "should mark only the changes in a contiguous block of differences as 'rejected'" do
              foo_diff = @foo_baseline.differences.find_by_resource_name('File[/etc/sudoers]')

              get :mark_as_reviewed, :id => @group.id, :date => @foo_report.time.strftime('%F'), :resource => 'File[/etc/sudoers]', :digest => foo_diff.resource_status_digest, :state => 'rejected'

              response.should be_success
              BaselineDifference.find_all_by_accepted(true).should_not be_present
              BaselineDifference.find_all_by_accepted(false).should =~ @expected
          end
        end

        describe 'where the action is not against the most recent change' do
          before do
            @foo_baselines[2].differences.find_by_resource_name('File[/etc/sudoers]').update_attributes! :resource_status_digest => ''
            @bar_baselines[4].differences.find_by_resource_name('File[/etc/sudoers]').update_attributes! :resource_status_digest => ''

            @expected = (@foo_baselines[3..-1] + @bar_baselines[3...4]).map do |bl|
              bl.differences.find_by_resource_name('File[/etc/sudoers]')
            end
          end

          it "should mark only the changes in a contiguous block of differences as 'accepted'" do
              foo_diff = @foo_baselines[3].differences.find_by_resource_name('File[/etc/sudoers]')

              get :mark_as_reviewed, :id => @group.id, :date => (@foo_report.time - 3.days).strftime('%F'), :resource => 'File[/etc/sudoers]', :digest => foo_diff.resource_status_digest, :state => 'accepted'

              response.should be_success
              BaselineDifference.find_all_by_accepted(true).should =~ @expected
              BaselineDifference.find_all_by_accepted(false).should_not be_present
          end

          it "should mark only the changes in a contiguous block of differences as 'rejected'" do
              foo_diff = @foo_baselines[3].differences.find_by_resource_name('File[/etc/sudoers]')

              get :mark_as_reviewed, :id => @group.id, :date => (@foo_report.time - 3.days).strftime('%F'), :resource => 'File[/etc/sudoers]', :digest => foo_diff.resource_status_digest, :state => 'rejected'

              response.should be_success
              BaselineDifference.find_all_by_accepted(true).should_not be_present
              BaselineDifference.find_all_by_accepted(false).should =~ @expected
          end
        end

        describe 'where the contiguous block contains previously accepted changes' do
          def do_setup(accepted)
            @old = 3.days.ago

            @foo_baselines[-2].
              differences.
              find_by_resource_name('File[/etc/sudoers]').
              update_attributes!(:resource_status_digest => '')
            @bar_baselines[-3].
              differences.
              find_by_resource_name('File[/etc/sudoers]').
              update_attributes!(:resource_status_digest => '')

            @foo_baselines[-3].
              differences.
              find_by_resource_name('File[/etc/sudoers]').
              update_attributes!(:decided_at => @old, :accepted => accepted)

            @bar_baselines[-4].
              differences.
              find_by_resource_name('File[/etc/sudoers]').
              update_attributes!(:decided_at => @old, :accepted => accepted)

            @foo_diffs = @foo_baselines.map do |bl|
              bl.differences.find_by_resource_name('File[/etc/sudoers]')
            end
            @bar_diffs = @bar_baselines.map do |bl|
              bl.differences.find_by_resource_name('File[/etc/sudoers]')
            end
          end

          it "should mark only the changes in a contiguous block of differences as 'accepted'" do
            do_setup(false)
            foo_diff = @foo_baseline.differences.
              find_by_resource_name('File[/etc/sudoers]')

            now = Time.now
            Time.stubs(:now).returns now

            get(:mark_as_reviewed, :id => @group.id,
                :date => @foo_report.time.strftime('%F'),
                :resource => 'File[/etc/sudoers]',
                :digest => foo_diff.resource_status_digest,
                :state => 'accepted')
            response.should be_success

            foo_diff.reload
            @new = now.to_i
            @old = @old.to_i

            @foo_diffs.map{|x| x.reload; [x.accepted, x.decided_at.to_i] }.
              should == [
                         [true, @new],
                         [true, @new],
                         [true, @new],
                         [false, @old],
                         [nil, 0],
                         [nil, 0]
                        ]

            @bar_diffs.map{|x| x.reload; [x.accepted, x.decided_at.to_i] }.
              should == [
                         [true, @new],
                         [true, @new],
                         [false, @old],
                         [nil, 0],
                         [nil, 0],
                         [nil, 0]
                        ]
          end

          it "should mark only the changes in a contiguous block of differences as 'rejected'" do
            do_setup(true)
            foo_diff = @foo_baseline.differences.
              find_by_resource_name('File[/etc/sudoers]')

            now = Time.now
            Time.stubs(:now).returns now
            get(:mark_as_reviewed, :id => @group.id,
                :date => @foo_report.time.strftime('%F'),
                :resource => 'File[/etc/sudoers]',
                :digest => foo_diff.resource_status_digest,
                :state => 'rejected')
            response.should be_success

            foo_diff.reload
            @new = now.to_i
            @old = @old.to_i

            @foo_diffs.map{|x| x.reload; [x.accepted, x.decided_at.to_i] }.
              should == [
                         [false, @new],
                         [false, @new],
                         [false, @new],
                         [true, @old],
                         [nil, 0],
                         [nil, 0]
                        ]

            @bar_diffs.map{|x| x.reload; [x.accepted, x.decided_at.to_i] }.
              should == [
                         [false, @new],
                         [false, @new],
                         [true, @old],
                         [nil, 0],
                         [nil, 0],
                         [nil, 0]
                        ]
          end
        end
      end
    end

    describe "#show_conflict" do
      it "should retrieve the values of the requested resource and end state for all nodes in the group" do
        foo_diff = @foo_baseline.differences.find_by_resource_name('File[/etc/sudoers]')
        bar_diff = @bar_baseline.differences.find_by_resource_name('File[/etc/sudoers]')

        foo_diff.update_attributes!(:decided_at => Time.now - 2.days, :accepted => true)

        get :show_conflict, :id => @group.id, :date => @foo_report.time.strftime('%F'), :digest => foo_diff.resource_status_digest, :resource => 'File[/etc/sudoers]'

        response.should be_success

        assigns[:multiple_baseline_states].should == [
          {
            :name => 'bar',
            :node_id => @bar_report.node.id,
            :diff_id => bar_diff.id,
            :accepted => nil,
            :can_change => true,
            :properties => {
              'owner' => {:baseline => nil, :inspected => '0'},
              'group' => {:baseline => nil, :inspected => '0'},
              'mode' => {:baseline => nil, :inspected => '440'}
            }
          },
          {
            :name => 'foo',
            :node_id => @foo_report.node.id,
            :diff_id => foo_diff.id,
            :accepted => true,
            :can_change => false,
            :properties => {
              'owner' => {:baseline => nil, :inspected => '0'},
              'group' => {:baseline => nil, :inspected => '0'},
              'mode' => {:baseline => nil, :inspected => '440'}
            }
          }
        ]
      end

      it "should properly handle the null digests" do
        foo_diff = @foo_baseline.differences.find_by_resource_name('File[/etc/sudoers]')
        bar_diff = @bar_baseline.differences.find_by_resource_name('File[/etc/sudoers]')

        foo_diff.update_attributes!(:resource_status_digest => nil)

        get :show_conflict, :id => @group.id, :date => @foo_report.time.strftime('%F'), :resource => 'File[/etc/sudoers]'

        response.should be_success

        assigns[:multiple_baseline_states].should == [
          {
            :name => 'foo',
            :node_id => @foo_report.node.id,
            :diff_id => foo_diff.id,
            :accepted => nil,
            :can_change => true,
            :properties => {
              'owner' => {:baseline => nil, :inspected => '0'},
              'group' => {:baseline => nil, :inspected => '0'},
              'mode' => {:baseline => nil,  :inspected => '440'}
            }
          }
        ]
      end

    end
  end
end
